/*
    parser.c
*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "parser.h"

static char *text;
static int textlen;
static int textpos;

/* precond: textpos < textlen */
static char *gettoken(char **blah) {
    static char tag[128];
    static char c;
    static int tagpos;
    static int strpos;
    static int initpos;
    static int strsize;

    tagpos = 0;
    strpos = 0;
    initpos = textpos;

    c = text[textpos++];
    
    if (isspace(c)) {          /* is a leading space */
        while (isspace(text[textpos++]) && textpos <= textlen);
        strsize = textpos - initpos;
        *blah = (char *) malloc(strsize);
        textpos = initpos;
        while (strpos < strsize) 
            (*blah)[strpos++] = text[textpos++];
        (*blah)[strpos-1] = '\0';
    }
    else if (c == '<') {            /* is a tag */
        while ( (tag[tagpos++] = text[textpos++]) != '>' && textpos <= textlen);
        tag[tagpos-1] = '\0';
        return tag;
    }
    else {                    /* is not a tag */
        while (text[textpos++] != '<' && textpos <= textlen);
        strsize = textpos - initpos;
        *blah = (char *) malloc(strsize);
        textpos = initpos;
        while (strpos < strsize)
            (*blah)[strpos++] = text[textpos++];
        (*blah)[strpos-1] = '\0';
        strpos -= 2;
        while (isspace((*blah)[strpos])) {
            (*blah)[strpos--] = '\0';
            textpos--;
        }
    }
    textpos--;
    return NULL;
}

static ENTITY *entitylist;
static int entitycount;
static int entitylistsize;

static void stackparse(ENTITY context, char *t);
static void irrelevant();
static void removespace(char *);
static void mergethings();
static void fixspaces();

static COLOR getcolor(char c)
{
    switch (c) {
    case 'r': return RED;
    case 'g': return GREEN;
    case 'b': return BLUE;
    case 'c': return CYAN;
    case 'm': return MAGENTA;
    case 'y': return YELLOW;
    case 'k': return BLACK;
    case 'w': return WHITE;
    default:  return UNDEF_C;
    }
}

ENTITY *parse(char *tx, int txlength)
{
    ENTITY root;
    text = tx; textlen = txlength; textpos = 0;

    root.text = NULL;
    root.bold = root.emph = root.italic =
        root.strong = root.tt = 0;
    root.underline = 0;
    root.size = 10;
    root.color = UNDEF_C;

    entitylistsize = 256;
    entitylist = malloc(sizeof(ENTITY) * entitylistsize);
    entitycount = 0;
    entitylist[0].text = NULL;

    stackparse(root, "");
    irrelevant();
    mergethings();
    fixspaces();

    return entitylist;
}

/* adds a new entity to the list of entities and returns a pointer to it */
static ENTITY *addentity()
{
    entitycount ++;
    if (entitylistsize <= entitycount) {
        ENTITY *newentlist = malloc(sizeof(ENTITY) * entitylistsize*2);

        memcpy(newentlist, entitylist, sizeof(ENTITY) * entitylistsize);

        free(entitylist);
        entitylist = newentlist;
        entitylistsize *= 2;
    }
    entitylist[entitycount].text = NULL;
    return &entitylist[entitycount-1];
}

static void stackparse(ENTITY context, char *t)
{
    char *str, *tag, opentag[128];
    strcpy(opentag, t);

    if (strcmp(t, "B") == 0) {
        context.bold = 1;
    }
    else if (strcmp(t, "EM") == 0) {
        if (context.strong == 0) {
            if (context.emph == 0)
                context.emph = 1;
            else
                context.emph = 0;
        }
    }
    else if (strcmp(t, "I") == 0) {
        context.italic = 1;
    }
    else if (strcmp(t, "PL") == 0) {
        context.bold = context.emph = context.italic =
            context.strong = context.tt = 0;
        context.underline = 0;
    }
    else if (strcmp(t, "S") == 0) {
        context.strong = 1;
        context.emph = 0;
    }
    else if (strcmp(t, "TT") == 0) {
        context.tt = 1;
    }
    else if (strcmp(t, "U") == 0) {
        if (context.underline < 3)
            context.underline ++;
    }
    else if (t[1] == 0 && isdigit(t[0])) {
        context.size = t[0] - '0';
    }
    else if (t[1] == 0 && islower(t[0])) {
        context.color = getcolor(t[0]);
    }

    while (textpos < textlen) {
        tag = gettoken(&str);
        if (tag) {
            if (tag[0] == '/')
                break;
            else
                stackparse(context, tag);
        }
        else {
            /* create a new entity with the current context */
            ENTITY *newent = addentity();
            *newent = context;
            newent->text = str;
        }
    }
    /* note: does not check whether it is the correct closing tag */
}

static int spaces(char *str)
{
    int i;
    for (i = 0; str[i]; i ++)
        if (!isspace(str[i])) return 0;
    return 1;
}

/* mark irrelevant attributes as irrelevant */
static void irrelevant()
{
    int i;
    for (i = 0; i < entitycount; i ++) {
        if (entitylist[i].strong)
            entitylist[i].emph = -1;
        if (spaces(entitylist[i].text)) {
            entitylist[i].emph = entitylist[i].italic =
                entitylist[i].bold = entitylist[i].strong = -1;
            if (entitylist[i].underline == 0)
                entitylist[i].color = IRREL_C;
        }
    }
}

int attrcmp(ENTITY *a, ENTITY *b)
{
    if (a->bold == b->bold && a->emph == b->emph &&
        a->italic == b->italic && a->strong == b->strong &&
        a->tt == b->tt && a->underline == b->underline &&
        a->size == b->size && a->color == b->color)
        return 1;
    else return 0;
}

/* merge adjacent entities with identical attributes */
static void mergethings(void)
{
    ENTITY *orig, *new;
    int l, i;
    orig = new = entitylist;
    entitylist[entitycount].text = NULL;
    while (orig->text) {
        for(l = 1; ; l ++) {
            if((orig+l)->text && attrcmp(orig, orig+l)) continue;
            else break;
        }
        if (l == 1) {
            *new = *orig;
            new ++; orig ++;
        }
        else {
            char *merged;
            int len, p;
            for (i = 0, len = 0; i < l; i ++)
                len += strlen((orig+i)->text);
            merged = malloc(len + 1);
            for (i = 0, p = 0; i < l; i ++) {
                strcpy(merged+p, (orig+i)->text);
                p += strlen((orig+i)->text);
                free((orig+i)->text);
                (orig+i)->text = NULL;
            }
            *new = *orig;
            new->text = merged;
            new ++; orig += l;
        }
    }
    new->text = NULL;
    entitycount = new - entitylist;
}

/* takes out all unnessecery whitespace in entitylist */
static void fixspaces(void)
{
    ENTITY *e;
    
    /* ie. all multiple spaces are reduced to one */
    
    for(e=entitylist;e->text;e++)
    {
        /* except if it's under tt */
        if(e->tt != 0) continue;
        
        removespace(e->text);
    }
}        

/* eliminates multiple spaces */
static void removespace(char *s)
{
    int p1, p2;
    p1 = p2 = 0;

    while (s[p1]) {
        if (isspace(s[p1])) {
            while (s[p1] && isspace(s[p1])) p1 ++;
            s[p2] = ' ';
            p2 ++;
        }
        else {
            s[p2] = s[p1];
            p2 ++; p1 ++;
        }
    }
    s[p2] = 0;
}
